#include "character.h"
#include <QtXml/qdom.h>
#include <QFile>
#include <QDebug>
#include "interface.h"
#include "game.h"
#include "player.h"
#include "monsters.h"

std::vector<CCharacter::SChampDBData>   CCharacter::dataBase;

//------------------------------------------------------
CCharacter::CCharacter()
{
    portrait = -1;
    spell[0] = 0;
    manaRegenTimer.set(MANA_REGEN_TIMER, true);
    healthRegenTimer.set(HEALTH_REGEN_TIMER, true);
}

//------------------------------------------------------
bool CCharacter::isDead()
{
    return (stats[eStatHealth].value <= 0);
}

//------------------------------------------------------
// initialize the character with values from the database
void CCharacter::fromDB(int num)
{
    SChampDBData&   db = dataBase[num];

    firstName = db.firstName;
    lastName = db.lastName;
    portrait = db.portrait;

    for (size_t i = 0; i < eStatCount; ++i)
    {
        stats[i].startValue = db.stats[i];
        stats[i].maxValue = db.stats[i];
        stats[i].value = db.stats[i];
    }
    stats[eStatLoad].value = 0;

    for (size_t i = 0; i < eBodyPartCount; ++i)
        bodyObjects[i].setType(0);

    for (size_t i = 0; i < db.objects.size(); ++i)
    {
        int pos = db.objects[i].pos;
        int type = db.objects[i].type;
        bodyObjects[pos].setType(type);
        objects.setDefaultParams(&bodyObjects[pos]);
    }

	// hidden skills XP
    for (size_t i = 4; i < eSkillCount; ++i)
    {
        if (db.skills[i] == 0)
            skills[i].experience = 0;
        else
            skills[i].experience = 250 << (db.skills[i] - 1);

        skills[i].startXP = skills[i].experience;
    }

	// main skills XP
    for (size_t i = 0; i < 4; ++i)
    {
        skills[i].experience = 0;

        for (size_t j = 0; j < 4; ++j)
            skills[i].experience += skills[4 * i + 4 + j].experience;

        skills[i].startXP = skills[i].experience;
    }

    // food, water
    food = 1500 + RANDOM(256);
    water = 1500 + RANDOM(256);
    foodTimer.set(FOOD_TIMER, true);
    waterTimer.set(WATER_TIMER, true);

    updateLoad();
}

//------------------------------------------------------
void CCharacter::readCharactersDB()
{
    // XML tags in the same order as EStatsNames
    static const char   statNames[][20] =
    {
        "health",
        "stamina",
        "mana",
        "strength",
        "dexterity",
        "wisdom",
        "vitality",
        "anti_magic",
        "anti_fire",
        "load"
    };

    QDomDocument doc;

    QFile f("databases/champions.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomNode    champNode = root.firstChild();

    while(!champNode.isNull())
    {
        if (champNode.isElement())
        {
            QDomElement champion = champNode.toElement();

            if (champion.tagName() == "champion")
            {
                SChampDBData   newChamp;

                std::string firstNameId = champion.attribute("firstName").toUtf8().constData();
                newChamp.firstName = interface.getText(firstNameId);
                std::string lastNameId = champion.attribute("lastName").toUtf8().constData();
                newChamp.lastName = interface.getText(lastNameId);

                for (int i = 0; i < eSkillCount; ++i)
                    newChamp.skills[i] = 0;

                QDomNode    champInfoNode = champion.firstChild();

                while(!champInfoNode.isNull())
                {
                    if (champInfoNode.isElement())
                    {
                        QDomElement champInfo = champInfoNode.toElement();

                        if (champInfo.tagName() == "portrait")
                        {
                            newChamp.portrait = champInfo.text().toInt();
                        }
                        else if (champInfo.tagName() == "object")
                        {
                            struct SDBObject    obj;

                            if (champInfo.attribute("pos") == "H")
                                obj.pos = eBodyPartHead;
                            else if (champInfo.attribute("pos") == "N")
                                obj.pos = eBodyPartNeck;
                            else if (champInfo.attribute("pos") == "T")
                                obj.pos = eBodyPartTorso;
                            else if (champInfo.attribute("pos") == "L")
                                obj.pos = eBodyPartLegs;
                            else if (champInfo.attribute("pos") == "F")
                                obj.pos = eBodyPartFeet;
                            else if (champInfo.attribute("pos") == "h")
                                obj.pos = eBodyPartRightHand;
                            else if (champInfo.attribute("pos") == "P")
                                obj.pos = eBodyPartPouch1;
                            else
                                obj.pos = champInfo.attribute("pos").toInt();

                            obj.type = champInfo.text().toInt();
                            newChamp.objects.push_back(obj);
                        }
                        else if (champInfo.tagName() == "skill")
                        {
                            int num = champInfo.attribute("num").toInt();
                            newChamp.skills[num] = champInfo.text().toInt();
                        }
                        else
                        {
                            for (size_t i = 0; i < eStatCount; ++i)
                            {
                                if (champInfo.tagName() == QString(statNames[i]))
                                    newChamp.stats[i] = champInfo.text().toInt();
                            }
                        }
                    }
                    champInfoNode = champInfoNode.nextSibling();
                }
                newChamp.stats[eStatLoad] *= 10;
                dataBase.push_back(newChamp);
            }
        }
        champNode = champNode.nextSibling();
    }
}

//------------------------------------------------------
void CCharacter::updateLoad()
{
    stats[eStatLoad].value = 0;

    for (int i = 0; i < eBodyPartCount; ++i)
        stats[eStatLoad].value += objects.getWeight(&bodyObjects[i]);

    if (&game.characters[interface.selectedChampion] == this)
        stats[eStatLoad].value += objects.getWeight(&mouse.mObjectInHand);
}

//------------------------------------------------------
int CCharacter::overload()
{
    int value = stats[eStatLoad].value;
    int max = stats[eStatLoad].maxValue;

    if (value > max)
        return 2;
    else if (value > (625 * max) / 1000)
        return 1;
    else
        return 0;
}

//------------------------------------------------------
void CCharacter::load(FILE* handle)
{
    // load the portrait
    fread(&portrait, sizeof(portrait), 1, handle);

    // load the name
    int size;
    char    c;
    fread(&size, sizeof(size), 1, handle);
    firstName.clear();
    for (int i = 0; i < size; ++i)
    {
        fread(&c, 1, 1, handle);
        firstName += c;
    }

    fread(&size, sizeof(size), 1, handle);
    lastName.clear();
    for (int i = 0; i < size; ++i)
    {
        fread(&c, 1, 1, handle);
        lastName += c;
    }

    // load the stats
    for (int i = 0; i < eStatCount; ++i)
    {
        fread(&stats[i].value, sizeof(stats[i].value), 1, handle);
        fread(&stats[i].maxValue, sizeof(stats[i].maxValue), 1, handle);
        fread(&stats[i].startValue, sizeof(stats[i].startValue), 1, handle);
    }

    // load the skills
    for (int i = 0; i < eSkillCount; ++i)
    {
        fread(&skills[i].experience, sizeof(skills[i].experience), 1, handle);
        fread(&skills[i].startXP, sizeof(skills[i].startXP), 1, handle);
    }

    // load the objects
    for (int i = 0; i < eBodyPartCount; ++i)
        bodyObjects[i].load(handle);

    // load the current spell
    fread(spell, 4, 1, handle);
    spell[4] = 0;

    // food and water
    fread(&food, sizeof(food), 1, handle);
    fread(&water, sizeof(water), 1, handle);
}

//------------------------------------------------------
void CCharacter::save(FILE* handle)
{
    // save the portrait
    fwrite(&portrait, sizeof(portrait), 1, handle);

    // save the name
    int size = firstName.size();
    fwrite(&size, sizeof(size), 1, handle);
    fwrite(firstName.c_str(), size, 1, handle);

    size = lastName.size();
    fwrite(&size, sizeof(size), 1, handle);
    fwrite(lastName.c_str(), size, 1, handle);

    // save the stats
    for (int i = 0; i < eStatCount; ++i)
    {
        fwrite(&stats[i].value, sizeof(stats[i].value), 1, handle);
        fwrite(&stats[i].maxValue, sizeof(stats[i].maxValue), 1, handle);
        fwrite(&stats[i].startValue, sizeof(stats[i].startValue), 1, handle);
    }

    // save the skills
    for (int i = 0; i < eSkillCount; ++i)
    {
        fwrite(&skills[i].experience, sizeof(skills[i].experience), 1, handle);
        fwrite(&skills[i].startXP, sizeof(skills[i].startXP), 1, handle);
    }

    // save the objects
    for (int i = 0; i < eBodyPartCount; ++i)
        bodyObjects[i].save(handle);

    // save the current spell
    fwrite(spell, 4, 1, handle);

    // save food and water
    fwrite(&food, sizeof(food), 1, handle);
    fwrite(&water, sizeof(water), 1, handle);
}

//------------------------------------------------------
void CCharacter::update()
{
    if (portrait == -1)
        return;

    // update torches
    if (bodyObjects[eBodyPartLeftHand].getType() == OBJECT_TYPE_TORCH)
    {
        CTimer  t;
        t.set(bodyObjects[eBodyPartLeftHand].getIntParam("Charges"), true);
        t.update();
        bodyObjects[eBodyPartLeftHand].setIntParam("Charges", t.get());
    }

    if (bodyObjects[eBodyPartRightHand].getType() == OBJECT_TYPE_TORCH)
    {
        CTimer  t;
        t.set(bodyObjects[eBodyPartRightHand].getIntParam("Charges"), true);
        t.update();
        bodyObjects[eBodyPartRightHand].setIntParam("Charges", t.get());
    }

    // update displayed damages
    damagesTimer.update();

    if (stats[eStatHealth].value != 0 && portrait != -1)
    {
        // mana regeneration
        if (manaRegenTimer.update() == true)
        {
            if (stats[eStatMana].value < stats[eStatMana].maxValue)
                stats[eStatMana].value++;

            if (interface.isSleeping() == false)
                manaRegenTimer.set(MANA_REGEN_TIMER, true);
            else
                manaRegenTimer.set(MANA_REGEN_TIMER_SLEEP, true);
        }

        // health regeneration
        if (healthRegenTimer.update() == true)
        {
            if (stats[eStatHealth].value < stats[eStatHealth].maxValue)
                stats[eStatHealth].value++;

            if (interface.isSleeping() == false)
                healthRegenTimer.set(HEALTH_REGEN_TIMER, true);
            else
                healthRegenTimer.set(HEALTH_REGEN_TIMER_SLEEP, true);
        }

        // food, water
        if (foodTimer.update() == true)
        {
            food--;
            if (food < MIN_FOOD)
                food = MIN_FOOD;

            if (interface.isSleeping() == false)
                foodTimer.set(FOOD_TIMER, true);
            else
                foodTimer.set(FOOD_TIMER_SLEEP, true);
        }

        if (waterTimer.update() == true)
        {
            water--;
            if (water < MIN_FOOD)
                water = MIN_FOOD;

            if (interface.isSleeping() == false)
                waterTimer.set(WATER_TIMER, true);
            else
                waterTimer.set(WATER_TIMER_SLEEP, true);
        }
    }
}

//------------------------------------------------------
void CCharacter::enterSleep()
{
    manaRegenTimer.set(MANA_REGEN_TIMER_SLEEP, true);
    healthRegenTimer.set(HEALTH_REGEN_TIMER_SLEEP, true);
    foodTimer.set(FOOD_TIMER_SLEEP, true);
    waterTimer.set(WATER_TIMER_SLEEP, true);
}

//------------------------------------------------------
void CCharacter::exitSleep()
{
    manaRegenTimer.set(MANA_REGEN_TIMER, true);
    healthRegenTimer.set(HEALTH_REGEN_TIMER, true);
    foodTimer.set(FOOD_TIMER, true);
    waterTimer.set(WATER_TIMER, true);
}

//------------------------------------------------------
int CCharacter::getDexterity()
{
    int dexterity = stats[eStatDexterity].value + RANDOM(8);

    // decrease dexterity depending on the load
    dexterity -= (dexterity / 2) * stats[eStatLoad].value / stats[eStatLoad].maxValue;

    // if we are sleeping...
    if (interface.isSleeping() == true)
        dexterity /= 2;

    // bound the value
    int min = 1 + RANDOM(8);
    int max = 100 - RANDOM(8);

    if (dexterity < min)
        dexterity = min;

    if (dexterity > max)
        dexterity = max;

    return dexterity;
}

//------------------------------------------------------
void CCharacter::hitChampion(int num, int damages)
{
    // TODO: wound, damage types

    // already dead or empty ?
    if (portrait == -1 || isDead() == true)
        return;

    displayedDamages = damages;
    damagesTimer.set(90, false);
    stats[eStatHealth].value -= damages;

    static CVec2    championPos[4][4] =
    {
        {{0, 0}, {1, 0}, {0, 1}, {1, 1}},   // up
        {{0, 1}, {0, 0}, {1, 1}, {1, 0}},   // left
        {{1, 1}, {0, 1}, {1, 0}, {0, 0}},   // down
        {{1, 0}, {1, 1}, {0, 0}, {0, 1}}    // right
    };

    // dead ?
    if (stats[eStatHealth].value <= 0)
    {
        stats[eStatHealth].value = 0;

        // drop all objects on the ground
        CVec2   stackPos = player.pos * 2 + championPos[player.dir][num];
        CObjectStack*   stack = map.addObjectsStack(stackPos, eWallSideMax);

        for (int i = 0; i < eBodyPartCount; ++i)
        {
            CObject&    obj = bodyObjects[i];

            if (obj.getType() != 0)
            {
                stack->addObject(obj);
                obj.setType(0);
            }
        }

        // drop the champion's bones
        CObject bones;
        bones.setType(OBJECT_TYPE_BONES);
        bones.setIntParam("Champion", num);
        stack->addObject(bones);
    }
}

//------------------------------------------------------
int CCharacter::getMeleeDamage(CVec2 monsterPos, CAttacks::SAttack& attack)
{
    int groupIndex = map.findMonsterGroupIndex(monsterPos);
    CMonsterGroup&  mapGroup = map.mMonsterGroups[groupIndex];
    int monsterType = mapGroup.getType();
    CMonsters::SMonsterData&    monsterData = monsters.monstersDatas[monsterType];
    int weaponType = bodyObjects[eBodyPartRightHand].getType();

    // do we hit ?
    int monsterDexterity = monsterData.dexterity + game.levelMultiplier() * 2 + RANDOM(32) - 16;
    int championDexterity = getDexterity();

    if (championDexterity > monsterDexterity ||
        ((RANDOM(75) + 1) <= attack.hitProbability) ||
        RANDOM(4) == 0)
    {
        int damages = 0;
        int monsterDefense = 0;

        // compute damages
        damages = getStrength(&bodyObjects[eBodyPartRightHand]);

        if (damages != 0)
        {
            damages += RANDOM(damages / 2 + 1);
            damages = (damages * attack.damages) / 32;
            monsterDefense = monsterData.defense + game.levelMultiplier() * 2 + RANDOM(32);

            if (weaponType == OBJECT_TYPE_DIAMOND_EDGE)
            {
                monsterDefense -= monsterDefense / 4;
            }
            else if (weaponType == OBJECT_TYPE_HARDCLEAVE)
            {
                monsterDefense -= monsterDefense / 8;
            }

            damages = damages - monsterDefense - RANDOM(32);
        }

        // last chance ?
        if (damages <= 1)
        {
            int damages2 = damages;
            damages = RANDOM(4);

            if (damages != 0)
            {
                damages++;
                damages2 += RANDOM(16);

                if (damages2 > 0 || RANDOM(2))
                {
                    damages += RANDOM(4);

                    if (RANDOM(4) == 0)
                        damages += MAX(0, damages2 + RANDOM(16));
                }
            }
        }

        // randomize the damages
        if (damages != 0)
        {
            damages /= 2;
            damages += RANDOM(damages) + RANDOM(4);
            damages += RANDOM(damages);
            damages /= 4;
            damages += RANDOM(4) + 1;
        }

        // skill bonus
        if (getSkillLevel((ESkillsNames)attack.skill) >= RANDOM(64))
        {
            damages += 10;
        }

        return damages;
    }
    return 0;
}

//------------------------------------------------------
int CCharacter::getStrength(CObject* obj)
{
    int strength = stats[eStatStrength].value + RANDOM(16);
    CObjects::CObjectInfo&    objInfo = objects.mObjectInfos[obj->getType()];

    // object weight
    int loadLimit = stats[eStatLoad].maxValue / 16;
    int loadLimit2 = loadLimit + (loadLimit - 12) / 2;
    int weight = objects.getWeight(obj);

    if (weight <= loadLimit)
    {
        strength += weight - 12;
    }
    else if (weight <= loadLimit2)
    {
        strength += (weight - loadLimit) / 2;
    }
    else
    {
        strength -= (weight - loadLimit2) * 2;
    }

    // weapon's damages
    strength += objInfo.damages;

    if (obj->getType() != 0)
    {
        int skill = 0;

        if (objInfo.weaponClass == WEAPON_CLASS_SWING || objInfo.weaponClass == WEAPON_CLASS_DAGGER_AND_AXES)
            skill = getSkillLevel(eSkillSwing) + 1;

        if (objInfo.weaponClass != WEAPON_CLASS_SWING && objInfo.weaponClass < WEAPON_CLASS_FIRST_BOW)
            skill += getSkillLevel(eSkillThrow) + 1;

        if (objInfo.weaponClass >= WEAPON_CLASS_FIRST_BOW && objInfo.weaponClass < WEAPON_CLASS_FIRST_MAGIC)
            skill += getSkillLevel(eSkillShoot) + 1;

        strength += skill * 2;
    }

    // bound the value
    strength /= 2;

    if (strength < 0)
        strength = 0;

    if (strength > 100)
        strength = 100;

    return strength;
}

//------------------------------------------------------
int CCharacter::getSkillLevel(ESkillsNames skillName)
{
    if (interface.isSleeping() == true)
        return 1;

    int exp = skills[skillName].experience;

    // Hidden skill
    if (skillName >= 4)
    {
        int mainSkill = (skillName - 4) / 4;
        exp = (exp + skills[mainSkill].experience) / 2;
    }

    // compute level from experience
    int level = 0;

    while (exp >= 500)
    {
        exp /= 2;
        level++;
    }

    return level;
}
